"""
ADiPy: Automatic Differentiation for Python

A simple tool for handling arbitrary order automatic
differentiation (AD).

See the examples at the bottom for some simple and advanced uses.

NOTES:

1. Arbitrary order differentiation only supported for univariate calculations
2. 1st-order differentiation supported for multivariate calculations (use
   numpy arrays).
3. Jacobian matrix generator
4. Univariate Taylor Series function generator
   
Copyright 2013: Abraham Lee

"""

from __future__ import division
import numpy as np
from math import factorial
import copy

class ad(object):

    def __init__(self, nom=None, der=None):
        """
        AD class constructor. A single nominal or a 1-d array of n nominal 
        values supported. If no value given for "der", then the default
        is "1" for a single nominal or an n-by-n identity matrix for a
        nominal array. Constants can be used by setting "der=0".
        """
        if nom is None:
            self.val = []
            self.der = []
        elif der is None:  # for constant with derivative 0
            self.val = nom
            try:
                nom[0]
            except:
                self.der = 1
            else:
                self.der = np.eye(len(nom))
        else:
            self.val = nom
            self.der = der
    
    def copy(self):
        return copy.copy(self)
        
    @property
    def nom(self):
        """
        Get the 0th-order value (i.e., the nominal value
        """
        return taylornominal(self)
    
    def d(self, n):
        """
        Get the nth derivative
        
        Example
        -------
        A 3rd-order differentiable object at 1.5::
        
            >>> x = adn(1.5, 3)
            >>> y = x**2
            >>> y.d(1)
            3.0
            >>> y.d(2)
            2.0
            >>> y.d(3)
            0.0
            
        """
        assert n>=0, 'Derivative order must not be negative.'
        if n==0:
            return self.nom
        else:
            derivs = taylorderivatives(self)
            assert len(derivs)>=n, \
                "You didn't track derivatives of order = {}".format(n)
            return derivs[n - 1]
    
    def __getitem__(self, idx):
        return ad(self.val[idx], self.der[idx])
    
    def __len__(self):
        return len(self.val)
        
    def __repr__(self):
        return 'ad' + repr((self.val, self.der))
    
    def __str__(self):
        return 'ad' + repr((self.nom, taylorderivatives(self)))
        
    def double(self):
        """
        Convert ad object to vector of doubles
        """
        return np.hstack((self.val*1.0, self.der*1.0))
        
    def __add__(self, other):
        if isinstance(other, ad):
            return ad(self.val + other.val, self.der + other.der)
        else:
            return ad(self.val + other, self.der)
    
    def __radd__(self, other):
        return self + other
        
    def __neg__(self):
        return -1*self
        
    def __sub__(self, other):
        return self + (-1)*other
    
    def __rsub__(self, other):
        return (-1)*self + other
    
    def __mul__(self, other):
        if isinstance(other, ad):
            return ad(self.val*other.val, 
                self.val*other.der + self.der*other.val)
        else:
            return ad(self.val*other, self.der*other)
    
    def __rmul__(self, other):
        return self*other
    
    def __div__(self, other):
        return self*(other**-1)
    
    def __rdiv__(self, other):
        return other*(self**-1)
    
    def __truediv__(self, other):
        return self.__div__(other)
    
    def __rtruediv__(self, other):
        return self.__rdiv__(other)
    
    def __pow__(self, other):
        if isinstance(other, ad):
            return ad(self.val**other.val, 
                other.val*self.val**(other.val - 1)*self.der + \
                self.val**other.val*log(self.val)*other.der)
        else:
            return ad(self.val**other, other*self.val**(other - 1)*self.der)
    
    def __rpow__(self, other):
        return ad(other**self.val, other**self.val*log(other)*self.der)
    
    def __abs__(self):
        if self==0:
            return ad(0*self.val, 0*self.der)
        else:
            return (self**2)**0.5
    
    def __eq__(self, other):
        if isinstance(other, ad):
            return self.nom==other.nom
        else:
            return self.nom==other
        
    def __lt__(self, other):
        if isinstance(other, ad):
            return self.nom<other.nom
        else:
            return self.nom<other
        
    def __le__(self, other):
        return self<other or self==other
    
    def __gt__(self, other):
        return not self<=other
    
    def __ge__(self, other):
        return not self<other
    
    def __ne__(self, other):
        return not self==other
    
    def __nonzero__(self):
        return self!=0
        
def adn(nom, order=1):
    """
    Construct an ad object that tracks derivatives up to ``order``.
    
    Parameters
    ----------
    nom : scalar
        The base, nominal value where the derivatives will be calculated at.
    
    Optional
    --------
    order : int
        The greatest order of derivatives to be tracked (Default: 1)
    """
    if order==1:
        return ad(nom)
    else:
        return ad(adn(nom, order - 1))

def taylornominal(u):
    """
    Collect the zeroth order term that would be used in a Taylor polynomial
    expansion (assumes any necessary calculations have been done prior).
    """
    if isinstance(u, ad):
        return taylornominal(u.val)
    else:
        return u

def taylorderivatives(u):
    """
    Collect all the derivative coefficients (unscaled) necessary for a Taylor
    polynomial expansion (assumes any necessary calculations have been done
    prior).
    """
    if isinstance(u, ad):
        order = 0
        tmp = u
        while hasattr(tmp, 'val'):
            tmp = tmp.val
            order += 1
        
        c = [0.0]*order
        for idx in xrange(order):
            tmp = u
            for v in xrange(order - idx - 1):
                tmp = tmp.val
            for d in xrange(idx + 1):
                if hasattr(tmp, 'der'):
                    tmp = tmp.der
                else:
                    break
            c[idx] = tmp
        return np.array(c)
    else:
        return np.ones_like(u)
    
def taylorcoef(u):
    """
    Collect all the scaled derivative coefficients necessary for a Taylor
    polynomial expansion (assumes any necessary calculations have been done
    prior).
    """
    if isinstance(u, ad):
        c = taylorderivatives(u)
        for idx in xrange(len(c)):
            c[idx] = c[idx]/factorial(idx + 1)
        return c
    else:
        return np.ones_like(u)

def taylorterms(u):
    """
    Collect all the coefficients necessary for a Taylor polynomial expansion,
    including the zeroth order term (assumes any necessary calculations have
    been done prior).
    """
    if isinstance(u, ad):
        return np.hstack((taylornominal(u), taylorcoef(u)))
    else:
        return np.array([u])

def taylorfunc(u, at=None):
    """
    Construct a univariate taylor polynomial expansion about a reference point.
    
    Parameters
    ----------
    u : an ad object
        The object that contains information about its derivatives
    at : scalar
        The point about which the series will be expanded (default: 0).
    
    Returns
    -------
    func : function
        The taylor series polynomial function that approximates ``u``, 
        expanded about the point ``at``.
    
    Example
    -------
    ::
    
        >>> from ad import series, taylorfunc
        >>> x = adn(3, 6)  # a sixth order derivative tracker, nominal=3
        >>> f = x*sin(x*x)
        >>> func = taylorfunc(f, 3)
        >>> func(3)
        1.2363554557252698
    """
    if at is None:
        at = 0  # assume it's expanded about the origin
        
    c = taylorterms(u)
    def approxfunc(x):
        try:
            x[0]
        except:
            return c[0] + np.sum([c[i]*(x - at)**i for i in xrange(1, len(c))])
        else:
            tmp = [c[0] + np.sum([c[i]*(xi - at)**i 
                for i in xrange(1, len(c))]) for xi in x]
            return np.array(tmp)
        
    return approxfunc
    
def exp(u):
    try:
        u[0]
    except:
        if isinstance(u, ad):
            return ad(exp(u.val), exp(u.val)*u.der)
        else:
            return np.exp(u)
    else:
        return [exp(ui) for ui in u]

def log(u):
    try:
        u[0]
    except:
        if isinstance(u, ad):
            return ad(log(u.val), 1/(u.val)*u.der)
        else:
            return np.log(u)
    else:
        return [log(ui) for ui in u]

def sqrt(u):
    try:
        u[0]
    except:
        if isinstance(u, ad):
            return ad(sqrt(u.val), u.der/(2*sqrt(u.val)))
        else:
            return np.sqrt(u)
    else:
        return [sqrt(ui) for ui in u]

def sin(u):
    try:
        u[0]
    except:
        if isinstance(u, ad):
            return ad(sin(u.val), cos(u.val)*u.der)
        else:
            return np.sin(u)
    else:
        return [sin(ui) for ui in u]

def cos(u):
    try:
        u[0]
    except:
        if isinstance(u, ad):
            return ad(cos(u.val), -sin(u.val)*u.der)
        else:
            return np.cos(u)
    else:
        return [cos(ui) for ui in u]
    
def tan(u):
    try:
        u[0]
    except:
        if isinstance(u, ad):
            return ad(tan(u.val), 1/(cos(u.val)**2)*u.der)
        else:
            return np.tan(u)
    else:
        return [tan(ui) for ui in u]
    
def asin(u):
    try:
        u[0]
    except:
        if isinstance(u, ad):
            return ad(asin(u.val), 1/sqrt(1 - u.val**2)*u.der)
        else:
            return np.arcsin(u)
    else:
        return [asin(ui) for ui in u]

def acos(u):
    try:
        u[0]
    except:
        if isinstance(u, ad):
            return ad(acos(u.val), -1/sqrt(1 - u.val**2)*u.der)
        else:
            return np.arccos(u)
    else:
        return [acos(ui) for ui in u]
    
def atan(u):
    try:
        u[0]
    except:
        if isinstance(u, ad):
            return ad(atan(u.val), 1/(1 + u.val**2)*u.der)
        else:
            return np.arctan(u)
    else:
        return [atan(ui) for ui in u]

def jacobian(deps):
    """
    Construct the Jacobian matrix of a set of AD objects that are dependent
    on other independent AD objects.
    
    Parameters
    ----------
    deps : array
        A list of objects that depend on AD objects.
        
    Returns
    -------
    jac : 2d-array
        A matrix where each row corresponds to each dependent AD object and 
        each column corresponds to each independent AD object.
        
    Example
    -------
    Separate ad objects::
    
        >>> x = ad(-1, np.array([1, 0, 0]))
        >>> y = ad(2.1, np.array([0, 1, 0]))
        >>> z = ad(0.25, np.array([0, 0, 1]))
        >>> u = x*y/z
        >>> v = y - z**x
        >>> w = sin(y**2/x)
        >>> jacobian([u, v, w])
        array([[  8.4       ,  -4.        ,  33.6       ],
               [  5.54517744,   1.        ,  16.        ],
               [  1.31330524,   1.25076689,   0.        ]])
    
    A single 3 ad object array::
    
        >>> x = ad(np.array([-1, 2.1, 0.25]))
        >>> y = [0]*3
        >>> y[0] = x[0]*x[1]/x[2]
        >>> y[1] = x[1] - x[2]**x[0]
        >>> y[2] = sin(x[1]**2/x[0])
        >>> jacobian(y)
        array([[  8.4       ,  -4.        ,  33.6       ],
               [  5.54517744,   1.        ,  16.        ],
               [  1.31330524,   1.25076689,   0.        ]])
    
    You can also break-out the individual components before using them::
    
        >>> x, y, z = ad(np.array([-1, 2.1, 0.25]))
        >>> x
        ad(-1.0, array([ 1.,  0.,  0.]))
        >>> y
        ad(2.1, array([ 0.,  1.,  0.]))
        >>> z
        ad(0.25, array([ 0.,  0.,  1.]))
       
    """
    assert all([isinstance(depi, ad) for depi in deps]), 'All inputs must be dependent on AD objects'
    return np.vstack([depi.d(1) for depi in deps])
    
def unite(ad_objs):
    """
    Unite multivariate AD objects into a single AD object
    
    Example
    -------
    
    ::
    
        >>> x = ad(np.array([-1, 2.1, 0.25]))
        >>> y = [0]*2
        >>> y[0] = x[0]*x[1]/x[2]
        >>> y[1] = x[1] - x[2]**x[0]
        # up to this point, y is a list, not an AD object, so y.d(1) won't work
        
        >>> y = unite(y)

        # now, we can do things like y.d(1)
        >>> y.d(1)
        array([[  8.4       ,  -4.        ,  33.6       ],
               [  5.54517744,   1.        ,  16.        ]])

        >>> jacobian(y)  # the result here is the same when y was a list object
        array([[  8.4       ,  -4.        ,  33.6       ],
               [  5.54517744,   1.        ,  16.        ]])
        
        
    """
    assert all([isinstance(adi, ad) for adi in ad_objs]), 'All inputs must be dependent on AD objects'
    try:
        ad_objs[0]
    except:
        return ad_objs
    else:
        ad_noms = np.array([adi.nom for adi in ad_objs])
        ad_ders = np.array([adi.der for adi in ad_objs])
        return ad(ad_noms, ad_ders)

if __name__=='__main__':
    def fdf(a):
        x = ad(a, 1)
        y = exp(-sqrt(x))*sin(x*log(1 + x**2))
        return y.double()
    
    def newtonfdf(a):
        delta = 1
        while abs(delta)>0.000001:
            fvec = fdf(a)
            delta = fvec[0]/fvec[1]
            a = a - delta
        return a
    
    print '\nnewtonfdf(5) should equal 4.8871:\n', newtonfdf(5)
    
    def fgradf(a0, v0, h0):
        a = ad(a0, np.array([1, 0, 0]))  # angle in degrees
        v = ad(v0, np.array([0, 1, 0]))  # velocity in ft/sec
        h = ad(h0, np.array([0, 0, 1]))  # height in ft
        rad = a*np.pi/180
        tana = tan(rad)
        vhor = (v*cos(rad))**2
        f = (vhor/32)*(tana + sqrt(tana**2 + 64*h/vhor))  # horizontal range
        return f.double()
    
    print '\nfgradf(20, 44, 9) should equal [56.0461, 1.0717, 1.9505, 1.4596]:\n', fgradf(20, 44, 9)
    
    def FJF(A):
        x = ad(A[0], np.array([1, 0, 0]))
        y = ad(A[1], np.array([0, 1, 0]))
        z = ad(A[2], np.array([0, 0, 1]))
        f1 = 3*x - cos(y*z) - 0.5
        f2 = x**2 - 81*(y + 0.1)**2 + sin(z) + 1.06
        f3 = exp(-x*y) + 20*z + (10*np.pi - 3)/3
        values = np.array([f1.double(), f2.double(), f3.double()])
        F = values[:, 0]
        J = values[:, 1:4]
        return (F, J)
        
    def newtonFJF(A):
        delta = 1
        while np.max(np.abs(delta))>0.000001:
            [F, J] = FJF(A)
            delta = np.linalg.solve(J, F)
            A = A - delta
        return A
        
    print '\nnewtonFJF([0.1, 0.1, -0.1]) should equal [0.5000, 0.0000, -0.5236]:\n', newtonFJF(np.array([0.1, 0.1, -0.1]))
    
    x = adn(3, 3)
    f = x*sin(x*x)
    print '\nf.d(3) should equal 495.9280:\n', f.d(3)



"""
This sub-module allows for the usage of several linear algebra routines that
are otherwise unavailable for use for automatic differentiation. Not all 
numpy.linalg routines have an equivalent here, but we're working on it.

Decompositions
    a. chol (Cholesky)
    b. qr (QR Factorization
    c. lu (LU Decomposition)

Solving equations and inverting matrices
    a. solve (Solve a linear matrix equation, or system of linear scalar eqs)
    b. lstsq (Solve a linear least squares problem)
    c. inv (Compute the (multiplicative) inverse of a matrix)

Calculating matrix norms
    a. norm (Frobenius norm, or 2-norm)
    
(c) 2013 by Abraham Lee <tisimst@gmail.com>.
Please send feature requests, bug reports, or feedback to this address.

This software is released under a dual license.  (1) The BSD license.
(2) Any other license, as long as it is obtained from the original
author.

"""

from __future__ import division
import numpy as np

__all__ = ['np', 'chol', 'qr', 'lu', 'solve', 'lstsq', 'inv', 'norm']

###############################################################################

def chol(A, side='lower'):
    """
    Cholesky decomposition of a symmetric, positive-definite matrix, defined
    by A = L*L.T = U.T*U, where L is a lower triangular matrix and U is an
    upper triangular matrix.
    
    Parameters
    ----------
    A : 2d-array
        The input matrix
    
    Optional
    --------
    side : str
        If 'lower' (default), then the lower triangular form of the
        decompostion is returned. If 'upper', then the upper triangular
        form of the decomposition is returned (the transpose of 'lower')
    
    Returns
    -------
    L : 2d-array
        The lower (or upper) triangular matrix that helps define the
        decomposition.
    
    Example
    -------
    Example 1::
    
        >>> A = [[25, 15, -5], 
        ...      [15, 18,  0], 
        ...      [-5,  0, 11]]
        ...
        >>> L = chol(A)
        >>> L
        array([[ 5.,  0.,  0.],
               [ 3.,  3.,  0.],
               [-1.,  1.,  3.]])
        >>> U = chol(A, 'upper')
        >>> U
        array([[ 5.,  3., -1.],
               [ 0.,  3.,  1.],
               [ 0.,  0.,  3.]])
        
    Example 2::
    
        >>> A = [[18, 22,  54,  42], 
        ...      [22, 70,  86,  62], 
        ...      [54, 86, 174, 134], 
        ...      [42, 62, 134, 106]]
        ...
        >>> L = chol(A)
        >>> L
        array([[  4.24264069,   0.        ,   0.        ,   0.        ],
               [  5.18544973,   6.5659052 ,   0.        ,   0.        ],
               [ 12.72792206,   3.0460385 ,   1.64974225,   0.        ],
               [  9.89949494,   1.62455386,   1.84971101,   1.39262125]])
           
    """
    A = np.array(A)
    assert A.shape[0]==A.shape[1], 'Input matrix must be square'
    
    L = [[0.0] * len(A) for _ in xrange(len(A))]
    for i in xrange(len(A)):
        for j in xrange(i+1):
            s = sum(L[i][k] * L[j][k] for k in xrange(j))
            L[i][j] = (A[i][i] - s)**0.5 if (i == j) else \
                      (1.0 / L[j][j] * (A[i][j] - s))
                          
    if side=='lower':
        return np.array(L)
    elif side=='upper':
        return np.array(L).T

###############################################################################

def qr(A):
    """
    QR Decomposition
    
    Parameters
    ----------
    A : 2d-array
        The input matrix (need not be square).
    
    Returns
    -------
    qm : 2d-array
        The orthogonal Q matrix from the decomposition
    rm : 2d-array
        The R matrix from the decomposition
    
    Example
    -------
    A square input matrix::
    
        >>> A = [[12, -51,   4], 
        ...      [ 6, 167, -68], 
        ...      [-4,  24, -41]]
        ...
        >>> q, r = qr(A)
        >>> q
        array([[-0.85714286,  0.39428571,  0.33142857],
               [-0.42857143, -0.90285714, -0.03428571],
               [ 0.28571429, -0.17142857,  0.94285714]])
        >>> r
        array([[ -1.40000000e+01,  -2.10000000e+01,   1.40000000e+01],
               [  5.97812398e-18,  -1.75000000e+02,   7.00000000e+01],
               [  4.47505281e-16,   0.00000000e+00,  -3.50000000e+01]])

    A non-square input matrix::
    
        >>> A = [[12, -51,   4], 
        ...      [ 6, 167, -68], 
        ...      [-4,  24, -41], 
        ...      [-1,   1,   0], 
        ...      [ 2,   0,   3]]
        ...
        >>> q, r = qr(A)
        >>> q
        array([[-0.84641474,  0.39129081, -0.34312406,  0.06613742, -0.09146206],
               [-0.42320737, -0.90408727,  0.02927016,  0.01737854, -0.04861045],
               [ 0.28213825, -0.17042055, -0.93285599, -0.02194202,  0.14371187],
               [ 0.07053456, -0.01404065,  0.00109937,  0.99740066,  0.00429488],
               [-0.14106912,  0.01665551,  0.10577161,  0.00585613,  0.98417487]])
        >>> r
        array([[ -1.41774469e+01,  -2.06666265e+01,   1.34015667e+01],
               [  3.31666807e-16,  -1.75042539e+02,   7.00803066e+01],
               [ -3.36067949e-16,   2.87087579e-15,   3.52015430e+01],
               [  9.46898347e-17,   5.05117109e-17,  -9.49761103e-17],
               [ -1.74918720e-16,  -3.80190411e-16,   8.88178420e-16]])
        >>> import numpy as np
        >>> np.all(np.dot(q, r) - A<1e-12)
        True
        
    """
    A = np.atleast_2d(A)
    m, n = A.shape
    qm = np.eye(m)
    rm = A.copy()
    for i in xrange(n-1 if m==n else n):
        x = getSubmatrix(rm, i, i, m, i)
        h = np.eye(m)
        h = setSubmatrix(h, i, i, householder(x))
        qm = np.dot(qm, h)
        rm = np.dot(h, rm)
    return qm, rm
    
###############################################################################

def lu(A):
    """
    Decomposes a nxn matrix A by PA=LU and returns L, U and P.
    
    Parameters
    ----------
    A : 2d-array
        The input matrix
   
    Returns
    -------
    L : 2d-array
        The lower-triangular matrix of the decomposition
    U : 2d-array
        The upper-triangular matrix of the decomposition
    P : 2d-array
        The pivoting matrix used in the decomposition
    
    Examples
    --------
    Example 1::
    
        >>> A = [[1, 3, 5],
        ...      [2, 4, 7],
        ...      [1, 1, 0]]
        ...
        >>> L, U, P = lu(A)
        >>> L
        array([[ 1. ,  0. ,  0. ],
               [ 0.5,  1. ,  0. ],
               [ 0.5, -1. ,  1. ]])
        >>> U
        array([[ 2. ,  4. ,  7. ],
               [ 0. ,  1. ,  1.5],
               [ 0. ,  0. , -2. ]])
        >>> P
        array([[ 0.,  1.,  0.],
               [ 1.,  0.,  0.],
               [ 0.,  0.,  1.]])
    
    Example 2::

        >>> A = [[11,  9, 24, 2], 
        ...      [ 1,  5,  2, 6], 
        ...      [ 3, 17, 18, 1], 
        ...      [ 2,  5,  7, 1]]
        ...
        >>> L, U, P = lu(A)
        >>> L
        array([[ 1.        ,  0.        ,  0.        ,  0.        ],
               [ 0.27272727,  1.        ,  0.        ,  0.        ],
               [ 0.09090909,  0.2875    ,  1.        ,  0.        ],
               [ 0.18181818,  0.23125   ,  0.00359712,  1.        ]])
        >>> U
        array([[ 11.        ,   9.        ,  24.        ,   2.        ],
               [  0.        ,  14.54545455,  11.45454545,   0.45454545],
               [  0.        ,   0.        ,  -3.475     ,   5.6875    ],
               [  0.        ,   0.        ,   0.        ,   0.51079137]])
        >>> P
        array([[ 1.,  0.,  0.,  0.],
               [ 0.,  0.,  1.,  0.],
               [ 0.,  1.,  0.,  0.],
               [ 0.,  0.,  0.,  1.]])
       
    """
    A = np.array(A)
    assert A.shape[0]==A.shape[1], 'Input matrix must be square'
    
    n = len(A)
    L = [[0.0]*n for i in xrange(n)]
    U = [[0.0]*n for i in xrange(n)]
    
    # Create the pivoting matrix for A
    P = [[float(i == j) for i in xrange(n)] for j in xrange(n)]
    for j in xrange(n):
        row = max(xrange(j, n), key=lambda i: A[i][j])
        if j != row:
            P[j], P[row] = P[row], P[j]
    
    A2 = np.dot(P, A)
    for j in xrange(n):
        L[j][j] = 1.0
        for i in xrange(j+1):
            s1 = sum(U[k][j] * L[i][k] for k in xrange(i))
            U[i][j] = A2[i][j] - s1
        for i in xrange(j, n):
            s2 = sum(U[k][j] * L[i][k] for k in xrange(j))
            L[i][j] = (A2[i][j] - s2) / U[j][j]
    return (np.array(L), np.array(U), np.array(P))
    
    
###############################################################################

def solve(A, b):
    """
    Solve a system of equations Ax = b by Gaussian elimination
    
    Parameters
    ----------
    A : 2d-array
        The LHS of the system of equations. The number of rows must not be less
        than the number of columns, but can be more.
    b : array-like
        The RHS of the system of equations (must have the same number of rows
        as ``A``. If more than one column is given, a solution is generated for
        each column
        
    Returns
    -------
    x : array-like
        The solution that satisifies the equality ``Ax==b``. If multiple ``b``
        are given, a solution column will be given satisfying each set.
    
    Example
    -------
    ::
    
        >>> A = [[1, 2, 1], [4, 6, 3], [9, 8, 2]]
        >>> b = [3, 2, 1]
        >>> solve(A, b)
        array([ -7.,  11., -12.])
        
    """
    try:
        A = np.matrix(A)
    except Exception:
        raise Exception('A must be a 2-dimensional array')
    assert A.shape[0]>=A.shape[1], 'A must not have less rows than columns'
    b = np.array(b)
    numout = 1 if b.ndim==1 else b.shape[1]
    assert A.shape[0]==b.shape[0], 'b must have the same number of rows as A'
    
    # Make the LHS array square if it isn't already and adjust RHS if needed
    if A.shape[0]>A.shape[1]:
        b = A.T*b
        A = A.T*A
        
    eqs = np.column_stack((A, b))
    n, m = eqs.shape
    
    # Forward substitution
    for k in range(n-1):
        # Pivot test
        p = k
        piv_el = abs(eqs[k, k])
        
        for i in range(k + 1, n):
            tmp = abs(eqs[i, k])
            if tmp>piv_el:
                piv_el = tmp
                p = i
        
        # Swap the kth and pth rows if a pivot element was found
        if p!=k:
            rtmp = np.array(eqs[p, :])
            eqs[p, :] = eqs[k, :]
            eqs[k, :] = rtmp
        
        for i in range(k + 1, n):
            f = 1.*eqs[i, k]/eqs[k, k]
            for j in range(k, m):
                eqs[i, j] -= f*eqs[k, j]
    
    # Backward substitution
    for k in range(1, n)[::-1]:
        for i in range(k)[::-1]:
            f = 1.*eqs[i, k]/eqs[k, k]
            for j in range(m)[::-1]:
                eqs[i, j] -= f*eqs[k, j]
    
    # Normalize
    for i in range(n):
        x = 1.*eqs[i, i]
        for j in range(m):
            eqs[i, j] = eqs[i, j]/x
    
    if numout==1:
        return np.array(eqs[:, -(m-n):]).ravel()
    else:
        return np.array(eqs[:, -(m-n):])

###############################################################################

def lstsq(A, b):
    """
    Solve Ax = b with the linear least squares method.
    
    Parameters
    ----------
    A : 2d-array
        The linear system matrix
    b : array
        The right-hand-side of the equation
    
    Returns
    -------
    x : array
        The solution to the system of equations
    
    Example
    -------
    (Taken from the NumPy lstsq example)::
    
        >>> x = np.array([0, 1, 2, 3])
        >>> y = np.array([-1, 0.2, 0.9, 2.1])
        >>> A = np.vstack([x, np.ones(len(x))]).T
        >>> A
        array([[ 0.,  1.],
               [ 1.,  1.],
               [ 2.,  1.],
               [ 3.,  1.]])
        >>> print lstsq(A, y)
        [ 1.   -0.95]
               
    """
    q, r = qr(A)
    n = r.shape[1]
    x = solveUpperTriangular(getSubmatrix(r, 0, 0, n - 1, n - 1), 
        np.dot(q.T, b))
    return x.ravel()

###############################################################################

def inv(A):
    """
    Calculate the multiplicative inverse of a matrix.
    
    Parameters
    ----------
    A : 2d-array
        The input matrix
        
    Returns
    -------
    Ainv : 2d-array
        The inverse of A
    
    Example
    -------
    ::

        >>> A = [[25, 15, -5], 
        ...      [15, 18,  0], 
        ...      [-5,  0, 11]]
        ...
        >>> Ainv = inv(A)
        >>> Ainv
        array([[ 0.09777778, -0.08148148,  0.04444444],
               [-0.08148148,  0.12345679, -0.03703704],
               [ 0.04444444, -0.03703704,  0.11111111]])
        >>> np.dot(Ainv, A)
        array([[  1.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  2.77555756e-16,   1.00000000e+00,   0.00000000e+00],
               [  0.00000000e+00,   1.11022302e-16,   1.00000000e+00]])
       
    """
    shape = np.atleast_2d(A).shape
    assert shape[0]==shape[1], 'Matrix must be square'
    return solve(A, np.eye(shape[0]))

###############################################################################

def norm(A):
    """
    Calculate the Frobenius norm of a matrix or vector, which is defined as
    the square root of the sum of the absolute squares of its elements.
    
    Parameters
    ----------
    A : array
        The input matrix or vector
        
    Returns
    -------
    N : scalar
        The inverse of A
    
    Example
    -------
    ::

        >>> A = [[25, 15, -5], 
        ...      [15, 18,  0], 
        ...      [-5,  0, 11]]
        ...
        >>> N = norm(A)
        >>> N
        XXXX
       
    """
    A = np.atleast_2d(A).ravel()
    return (np.sum(np.abs(A)**2))**0.5
    
###############################################################################
#
#
#    All functions below are utility functions for some of the above
#    linear algebra functions. They should NOT be used directly.
#
#
###############################################################################

def ematrix(m, n, x, i, j):
    a = np.empty((m, n))
    a[:, :] = 0.0
    a[i, j] = x
    return a
    
def unitVector(n):
    return ematrix(n, 1, 1, 0, 0)

def signValue(r):
    if r<0:
        return -1
    elif r==0:
        return 0
    else:
        return 1

def householder(A):
    m = len(A)
    u = A - np.sqrt(np.dot(A.T, A)[0, 0])*unitVector(m)
    v = u
    beta = 2/np.dot(v.T, v)
    return np.eye(m) - beta*(np.dot(v, v.T).T)

def getSubmatrix(obj, i1, j1, i2, j2):
    """
    i1, j1, i2, j2 are inclusive indices
    """
    return obj[i1:i2 + 1, j1:j2 + 1]

def setSubmatrix(obj, i1, j1, subobj):
    """
    i1, j1 are the top left corner indices where subobj will be inserted
    """
    m, n = np.atleast_2d(subobj).shape
    obj[i1:i1+m, j1:j1+n] = subobj
    return obj

def solveUpperTriangular(r, b):
    r = np.atleast_2d(r)
    n = r.shape[1]
    
    x = np.zeros((n, 1))
    for k in xrange(n - 1, -1, -1):
        idx = min(n - 1, k)
        x[k, 0] = (b[k] - np.dot(getSubmatrix(r, k, idx, k, n - 1),
            getSubmatrix(x, idx, 0, n - 1, 0)))/r[k, k]
    return x

def polyfit(x, y, n):
    """
    Fit a polynomial of order n to data arrays x and y
    
    Parameters
    ----------
    x : array
        A data array that defines the first dimension coordinates.
    y : array
        A data array that defines the second dimension coordinates.
    n : int
        The order of the polynomial (e.g., 1 for linear, 2 for quadratic, etc.)
    
    Returns
    -------
    b : array
        The polynomial coefficients, in ascending order.
    
    Example
    -------
    ::
    
        >>> x = range(11)
        >>> y = [1, 6, 17, 34, 57, 86, 121, 162, 209, 262, 321]
        >>> polyfit(x, y, 2)
        array([ 1.,  2.,  3.])
        
    """
    a = np.empty((len(x), n + 1))
    for i in xrange(a.shape[0]):
        for j in xrange(a.shape[1]):
            a[i, j] = 1 if j==0 else x[i]**j
    return lstsq(a, y)
